/*
 *	Forwarding database
 *	Linux ethernet bridge
 *
 *	Authors:
 *	Lennert Buytenhek		<buytenh@gnu.org>
 *
 *	$Id: br_fdb.c,v 1.1.1.1 2002/03/18 15:12:19 jharrell Exp $
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 */

//-------------------------------------------------------------------------------------
// Copyright 2004, Texas Instruments Incorporated
//
// This program has been modified from its original operation by Texas Instruments
// to do the following:
// 
// Performance optimizations
//
// THIS MODIFIED SOFTWARE AND DOCUMENTATION ARE PROVIDED
// "AS IS," AND TEXAS INSTRUMENTS MAKES NO REPRESENTATIONS 
// OR WARRENTIES, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED 
// TO, WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY 
// PARTICULAR PURPOSE OR THAT THE USE OF THE SOFTWARE OR 
// DOCUMENTATION WILL NOT INFRINGE ANY THIRD PARTY PATENTS, 
// COPYRIGHTS, TRADEMARKS OR OTHER RIGHTS. 
// See The GNU General Public License for more details.
//
// These changes are covered under version 2 of the GNU General Public License, 
// dated June 1991.
//-------------------------------------------------------------------------------------

/* The definition controls the orignal Linux Code and new optimized code.
 * Define this MACRO to switch back to the orignal Linux Code. */
#ifndef TI_SLOW_PATH

#include <linux/kernel.h>
#include <linux/spinlock.h>
#include <linux/if_bridge.h>
#include <asm/atomic.h>
#include <asm/uaccess.h>
#include "br_private.h"
/* frank,040202 */
#ifdef ASUS_IGMP_SNOOPING
#include <linux/inetdevice.h>
#include <linux/ip.h>
#include <asm/checksum.h>
#endif

#if 0
static __inline__ unsigned long __timeout(struct net_bridge *br)
{
	unsigned long timeout;

	timeout = jiffies - br->ageing_time;
	if (br->topology_change)
		timeout = jiffies - br->forward_delay;

	return timeout;
}

static __inline__ int has_expired(struct net_bridge *br, struct net_bridge_fdb_entry *fdb)
{
	if (!fdb->is_static &&
	    time_before_eq(fdb->ageing_timer, __timeout(br)))
		return 1;

	return 0;
}
#endif

static __inline__ void copy_fdb(struct __fdb_entry *ent, struct net_bridge_fdb_entry *f)
{
	memset(ent, 0, sizeof(struct __fdb_entry));
	memcpy(ent->mac_addr, f->addr.addr, ETH_ALEN);
	ent->port_no = f->dst?f->dst->port_no:0;
	ent->is_local = f->is_local;
	ent->ageing_timer_value = 0;
	if (!f->is_static)
		ent->ageing_timer_value = jiffies - f->ageing_timer;
}

#if 0
static __inline__ int br_mac_hash(unsigned char *mac)
{
	unsigned long x;

	x = mac[0];
	x = (x << 2) ^ mac[1];
	x = (x << 2) ^ mac[2];
	x = (x << 2) ^ mac[3];
	x = (x << 2) ^ mac[4];
	x = (x << 2) ^ mac[5];

	x ^= x >> 8;

	return x & (BR_HASH_SIZE - 1);
}
#endif

static __inline__ void __hash_link(struct net_bridge *br,
				   struct net_bridge_fdb_entry *ent,
				   int hash)
{
	ent->next_hash = br->hash[hash];
	if (ent->next_hash != NULL)
		ent->next_hash->pprev_hash = &ent->next_hash;
	br->hash[hash] = ent;
	ent->pprev_hash = &br->hash[hash];
}

static __inline__ void __hash_unlink(struct net_bridge_fdb_entry *ent)
{
	*(ent->pprev_hash) = ent->next_hash;
	if (ent->next_hash != NULL)
		ent->next_hash->pprev_hash = ent->pprev_hash;
	ent->next_hash = NULL;
	ent->pprev_hash = NULL;
}

/**************************************************************************
 * FUNCTION NAME : maccmp
 **************************************************************************
 * DESCRIPTION   :
 *  The function compares the mac addresses. This used to be a memcmp which
 *  used to compare the address in 6 iterations, using unsigned short 
 *  comparisons the work can now be done in 3 iterations.
 *
 * RETURNS       :
 *      0 - If MAC Addresses are the same
 *      1 - If they are not the same.
 ***************************************************************************/
static __inline__ int maccmp(unsigned char* addr1, unsigned char* addr2)
{
    if ( (*((unsigned short *)addr1) == *((unsigned short *)addr2))     &&
         (*((unsigned short *)(addr1+2)) == *((unsigned short *)(addr2+2))) &&
         (*((unsigned short *)(addr1+4)) == *((unsigned short *)(addr2+4))) )
    {
        return 0;
    }
    return 1;
}

/* These function have been moved to the BRIDGE FAST Path. */
#if 0
struct net_bridge_fdb_entry *br_fdb_get(struct net_bridge *br, unsigned char *addr)
{
	struct net_bridge_fdb_entry *fdb;

	read_lock_bh(&br->hash_lock);
	fdb = br->hash[br_mac_hash(addr)];
	while (fdb != NULL)
    {
        /* NSP Optimizations: Use the function maccmp instead of memcmp. */
        if (!maccmp(fdb->addr.addr, addr))
        {
			if (!has_expired(br, fdb)) 
            {
				atomic_inc(&fdb->use_count);
				read_unlock_bh(&br->hash_lock);
				return fdb;
			}

			read_unlock_bh(&br->hash_lock);
			return NULL;
		}

		fdb = fdb->next_hash;
	}

	read_unlock_bh(&br->hash_lock);
	return NULL;
}

void br_fdb_put(struct net_bridge_fdb_entry *ent)
{
	if (atomic_dec_and_test(&ent->use_count))
		kfree(ent);
}
#endif

void br_fdb_changeaddr(struct net_bridge_port *p, unsigned char *newaddr)
{
	struct net_bridge *br;
	int i;

	br = p->br;
	write_lock_bh(&br->hash_lock);
	for (i=0;i<BR_HASH_SIZE;i++) {
		struct net_bridge_fdb_entry *f;

		f = br->hash[i];
		while (f != NULL) {
			if (f->dst == p && f->is_local) {
				__hash_unlink(f);
				memcpy(f->addr.addr, newaddr, ETH_ALEN);
				__hash_link(br, f, br_mac_hash(newaddr));
				write_unlock_bh(&br->hash_lock);
				return;
			}
			f = f->next_hash;
		}
	}
	write_unlock_bh(&br->hash_lock);
}

void br_fdb_cleanup(struct net_bridge *br)
{
	int i;
	unsigned long timeout;

	timeout = __timeout(br);

	write_lock_bh(&br->hash_lock);
	for (i=0;i<BR_HASH_SIZE;i++) {
		struct net_bridge_fdb_entry *f;

		f = br->hash[i];
		while (f != NULL) {
			struct net_bridge_fdb_entry *g;

			g = f->next_hash;
			if (!f->is_static &&
			    time_before_eq(f->ageing_timer, timeout)) {
				__hash_unlink(f);
				br_fdb_put(f);
			}
			f = g;
		}
	}
	write_unlock_bh(&br->hash_lock);
}

void br_fdb_delete_by_port(struct net_bridge *br, struct net_bridge_port *p)
{
	int i;

	write_lock_bh(&br->hash_lock);
	for (i=0;i<BR_HASH_SIZE;i++) {
		struct net_bridge_fdb_entry *f;

		f = br->hash[i];
		while (f != NULL) {
			struct net_bridge_fdb_entry *g;

			g = f->next_hash;
			if (f->dst == p) {
				__hash_unlink(f);
				br_fdb_put(f);
			}
			f = g;
		}
	}
	write_unlock_bh(&br->hash_lock);
}

int br_fdb_get_entries(struct net_bridge *br,
		       unsigned char *_buf,
		       int maxnum,
		       int offset)
{
	int i;
	int num;
	struct __fdb_entry *walk;

	num = 0;
	walk = (struct __fdb_entry *)_buf;

	read_lock_bh(&br->hash_lock);
	for (i=0;i<BR_HASH_SIZE;i++) {
		struct net_bridge_fdb_entry *f;

		f = br->hash[i];
		while (f != NULL && num < maxnum) {
			struct __fdb_entry ent;
			int err;
			struct net_bridge_fdb_entry *g;
			struct net_bridge_fdb_entry **pp; 

			if (has_expired(br, f)) {
				f = f->next_hash;
				continue;
			}

			if (offset) {
				offset--;
				f = f->next_hash;
				continue;
			}

			copy_fdb(&ent, f);

			atomic_inc(&f->use_count);
			read_unlock_bh(&br->hash_lock);
			err = copy_to_user(walk, &ent, sizeof(struct __fdb_entry));
			read_lock_bh(&br->hash_lock);

			g = f->next_hash;
			pp = f->pprev_hash;
			br_fdb_put(f);

			if (err)
				goto out_fault;

			if (g == NULL && pp == NULL)
				goto out_disappeared;

			num++;
			walk++;

			f = g;
		}
	}

 out:
	read_unlock_bh(&br->hash_lock);
	return num;

 out_disappeared:
	num = -EAGAIN;
	goto out;

 out_fault:
	num = -EFAULT;
	goto out;
}

static __inline__ void __fdb_possibly_replace(struct net_bridge_fdb_entry *fdb,
					      struct net_bridge_port *source,
					      int is_local)
{
	if (!fdb->is_static || is_local) {
		fdb->dst = source;
		fdb->is_local = is_local;
		fdb->is_static = is_local;
		fdb->ageing_timer = jiffies;
	}
}

void br_fdb_insert(struct net_bridge *br,
		   struct net_bridge_port *source,
		   unsigned char *addr,
		   int is_local)
{
	struct net_bridge_fdb_entry *fdb;
	int hash;

	hash = br_mac_hash(addr);

	write_lock_bh(&br->hash_lock);
	fdb = br->hash[hash];
	while (fdb != NULL)
    {
        /* NSP Optimizations: Use maccmp instead of the slow memcmp. */
#if 0
		if (!memcmp(fdb->addr.addr, addr, ETH_ALEN))
#else
        if (!maccmp(fdb->addr.addr, addr))
#endif
        {

			__fdb_possibly_replace(fdb, source, is_local);
			write_unlock_bh(&br->hash_lock);
			return;
		}

		fdb = fdb->next_hash;
	}

	fdb = kmalloc(sizeof(*fdb), GFP_ATOMIC);
	if (fdb == NULL) {

		write_unlock_bh(&br->hash_lock);
		return;
	}

	memcpy(fdb->addr.addr, addr, ETH_ALEN);
	atomic_set(&fdb->use_count, 1);
	fdb->dst = source;
	fdb->is_local = is_local;
	fdb->is_static = is_local;
	fdb->ageing_timer = jiffies;

	__hash_link(br, fdb, hash);

	write_unlock_bh(&br->hash_lock);
}

/* frank,040202 */
#ifdef ASUS_IGMP_SNOOPING
static __inline__ void __mhash_link(struct net_bridge *br,
				   struct net_bridge_mdb_entry *ent,
				   int mhash)
{
	ent->next_hash = br->mhash[mhash];
	if (ent->next_hash != NULL)
		ent->next_hash->pprev_hash = &ent->next_hash;
	br->mhash[mhash] = ent;
	ent->pprev_hash = &br->mhash[mhash];
}

static __inline__ void __mhash_unlink(struct net_bridge_mdb_entry *ent)
{
	*(ent->pprev_hash) = ent->next_hash;
	if (ent->next_hash != NULL)
		ent->next_hash->pprev_hash = ent->pprev_hash;
	ent->next_hash = NULL;
	ent->pprev_hash = NULL;
}

struct net_bridge_mdb_entry *br_mdb_get(struct net_bridge *br, unsigned char *ip)
{
	struct net_bridge_mdb_entry *mdb;
	unsigned char addr[6]={0,0};

	read_lock_bh(&br->mhash_lock);
	memcpy(&addr[2],ip,4);
	mdb = br->mhash[br_mac_hash(addr)];
	while (mdb != NULL) {
		if (mdb->addr==*((unsigned int *)ip)) {				
			read_unlock_bh(&br->mhash_lock);
			return mdb;
		}
		mdb = mdb->next_hash;
	}
	read_unlock_bh(&br->mhash_lock);
	return NULL;
}
void br_mdb_insert(struct net_bridge *br,
		   struct net_bridge_port *source,
		   unsigned char *ip)
{
	struct net_bridge_mdb_entry *mdb;
	struct net_bridge_mport	 *mdst;
	unsigned char addr[6]={0,0};
	int mhash;

	memcpy(&addr[2],ip,4);
	mhash = br_mac_hash(addr);

	write_lock_bh(&br->mhash_lock);
	mdb = br->mhash[mhash];
	while (mdb != NULL) {
		if (mdb->addr==*((unsigned int *) ip)) {
			/* check port */
			mdst=mdb->mdst;
			while(mdst != NULL)
				{
				if(mdst->dst == source)
					{
					/* old ones,reset aging timer */
					mdst->ageing_timer=jiffies;
#ifdef ASUS_IGMP_SNOOPING_DEBUG					
					printk("[%u]reset timer from [%s] for [01:00:5E:%02X:%02X:%02X]\n",jiffies/100,source->dev->name,addr[3],addr[4],addr[5]);
#endif
					break;
					}
				else
					mdst=mdst->next_port;
				}
			/* not found bridge port in mfdb,insert port to mdb */
			if(mdst == NULL)
				{
				mdst=kmalloc(sizeof(*mdst), GFP_ATOMIC);			
				if(mdst == NULL)
					{
					write_unlock_bh(&br->mhash_lock);
					printk("can't allocate mem for mdst\n");
					return;
					}
				mdst->ageing_timer=jiffies;
				mdst->dst=source;
				//nick, 041121, if the new mdst is first one to be add, nill it's next_port.
				if (mdb->mdst == NULL)
					mdst->next_port=NULL;
				else
					mdst->next_port=mdb->mdst;

				mdb->mdst=mdst;
#ifdef ASUS_IGMP_SNOOPING_DEBUG					
				printk("[%u]new mdst  from [%s] for [01:00:5E:%02X:%02X:%02X]\n",jiffies/100,source->dev->name,addr[3],addr[4],addr[5]);
#endif
				}
			write_unlock_bh(&br->mhash_lock);
			return;
		}

		mdb = mdb->next_hash;
	}

	mdb = kmalloc(sizeof(*mdb), GFP_ATOMIC);
	mdst=kmalloc(sizeof(*mdst), GFP_ATOMIC);
	
	if (mdb == NULL || mdst == NULL) {
		if(mdb != NULL)
		{
			printk("can't allocate mem for mdst\n");
			kfree(mdb);
		}
		if(mdst != NULL)
		{
			printk("can't allocate mem for mdb\n");
			kfree(mdst);
		}
		write_unlock_bh(&br->mhash_lock);
			
		return;
	}
	mdst->ageing_timer=jiffies;
	mdst->next_port=NULL;
	mdst->dst=source;
	mdb->addr=*((unsigned int *)ip);
	/* copy igmp packet */	
		

	atomic_set(&mdb->use_count, 1);
	mdb->mdst = mdst;

	__mhash_link(br, mdb, mhash);
#ifdef ASUS_IGMP_SNOOPING_DEBUG					
	printk("[%u]mew mdb  from [%s] for [01:00:5E:%02X:%02X:%02X]\n",jiffies/100,source->dev->name,addr[3],addr[4],addr[5]);
#endif
	write_unlock_bh(&br->mhash_lock);
}

void br_mdb_leave_group(struct net_bridge *br,
		   struct net_bridge_port *source,
		   unsigned char *group)
{
	struct net_bridge_mdb_entry *mdb;
	struct net_bridge_mport	 *mdst,*before;
	unsigned char addr[6]={0x00,0x00};

	memcpy(&addr[2],group,4);
#ifdef ASUS_IGMP_SNOOPING_DEBUG					
	printk("Receive IGMP Leave Group[01:00:5e:%02X:%02X:%02X] from port[%s]\n",addr[3],addr[4],addr[5],source->dev->name);
#endif
    return;	
	read_lock_bh(&br->mhash_lock);
	mdb = br->mhash[br_mac_hash(addr)];
	while (mdb != NULL) {
		if (mdb->addr==*((unsigned int *)group)) {
			mdst=mdb->mdst;
			before=mdst;
			while(mdst != NULL)
				{
				if(mdst->dst == source)
					{
					if(mdst == mdb->mdst) /* remove first */
						{
						mdb->mdst=mdst->next_port;
						kfree(mdst);
						mdst=mdb->mdst;
						before=mdst;
						}
					else
						{
						before->next_port=mdst->next_port;
						kfree(mdst);
						mdst=before->next_port;
						}
#ifdef ASUS_IGMP_SNOOPING_DEBUG					
					printk("Find Entry and delete\n\n");
#endif
					break;
					}
				else
					{
					before=mdst;
					mdst=mdst->next_port;
					}
				}
			if(mdb->mdst == NULL)
				{
				__mhash_unlink(mdb);
					{
#ifdef ASUS_IGMP_SNOOPING_DEBUG					
					printk("clean MDB\n");
#endif
					kfree(mdb);
					mdb=NULL;
					}
				}
				
			read_unlock_bh(&br->mhash_lock);
			return;
		}

		mdb = mdb->next_hash;
	}

	read_unlock_bh(&br->mhash_lock);
}

static unsigned char igmp_query_pkt[60]={
	0x01,0x00,0x5e,0x00,0x00,0x01,0x00,0x00,0x00,0x00,0x00,0x00,0x08,0x00,0x45,0xc0,
	0x00,0x1c,0xe5,0xb4,0x00,0x00,0x01,0x02,0xea,0xc0,0xc0,0xa8,0x48,0x01,0xe0,0x00,
	0x00,0x01,0x11,0x64,0xee,0x9b,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00};
void br_send_igmp_query(struct net_bridge_port *port)
{
	struct net_device *dev;
	struct sk_buff *skb;
	int size;
	struct in_ifaddr *ifa=NULL;
	struct iphdr	*iph;

	/* get ip address of br */
	if(port->br->dev.ip_ptr != NULL)
		ifa = ((struct in_device*)(port->br->dev.ip_ptr))->ifa_list;
	
	
	size = 60;
	dev = port->dev;

	if ((skb = dev_alloc_skb(size)) == NULL) {
		printk(KERN_INFO "br: memory squeeze!\n");
		return;
	}
#ifdef ASUS_IGMP_SNOOPING_DEBUG					
	printk("[%u]send IGMP Query:port[%s] \n\n",jiffies/HZ,port->dev->name);
#endif
	skb->dev = dev;
	skb->protocol = htons(ETH_P_IP);
	skb->mac.raw = skb_put(skb, size);
	memcpy(skb->mac.raw, igmp_query_pkt, size);
	memcpy(skb->mac.raw+ETH_ALEN, dev->dev_addr, ETH_ALEN);

	if(ifa != NULL)
		{
		(unsigned char *)iph=skb->mac.raw+14;
		iph->saddr=ifa->ifa_address;
		iph->check=0;
		iph->check=ip_fast_csum((unsigned char *) iph, iph->ihl);
		}
	dev_queue_xmit(skb);


}

int br_mdb_get_entries(struct net_bridge *br,
		       unsigned char *_buf,
		       int maxnum,
		       int offset)
{
	struct net_bridge_mdb_entry *mdb;
	struct net_bridge_mport	 *mdst;
	struct __mdb_entry *data,ent;
	int index=0,i;

	data=(struct __mdb_entry *)_buf;

	read_lock_bh(&br->mhash_lock);
		
	for(i=0;i<BR_HASH_SIZE;i++)
		{
		mdb = br->mhash[i];
		while (mdb != NULL) 
			{
			mdst=mdb->mdst;
			while(mdst != NULL) 
				{
				ent.ip_addr=mdb->addr;
				ent.ageing_timer_value=(jiffies-mdst->ageing_timer)/HZ;
				ent.port_no=mdst->dst->port_no;
				copy_to_user(&data[index], &ent, sizeof(struct __mdb_entry));
				index++;
				if(index >= maxnum)
					{
					read_unlock_bh(&br->mhash_lock);
					return index;
					}
				mdst=mdst->next_port;
				}				
			mdb = mdb->next_hash;
			}
		}
	read_unlock_bh(&br->mhash_lock);
	return index;
}

void br_mdb_cleanup(struct net_bridge *br)
{
	struct net_bridge_mdb_entry *mdb,*pre_mdb;
	struct net_bridge_mport	 *mdst,*pre_mdst;
	int i;

	return;
	read_lock_bh(&br->mhash_lock);
	for(i=0;i<BR_HASH_SIZE;i++)
		{
		mdb = br->mhash[i];
		while (mdb != NULL) 
			{
			mdst=mdb->mdst;
			while(mdst != NULL) 
				{
				pre_mdst=mdst;
				mdst=mdst->next_port;
				kfree(pre_mdst);
				}		
			pre_mdb=mdb;
			mdb=mdb->next_hash;
			__mhash_unlink(pre_mdb);
			kfree(pre_mdb);
			}
		}
	read_unlock_bh(&br->mhash_lock);
}

void br_mdb_delete_by_port(struct net_bridge *br, struct net_bridge_port *p)
{
	struct net_bridge_mdb_entry *mdb,*pre_mdb;
	struct net_bridge_mport	 *mdst,*before;
	int i;

	read_lock_bh(&br->mhash_lock);
	for(i=0;i<BR_HASH_SIZE;i++)
		{
		mdb = br->mhash[i];
		while (mdb != NULL) 
			{
			mdst=mdb->mdst;
			before=mdst;
			while(mdst != NULL) 
				{
				if(mdst->dst == p)
					{
#ifdef ASUS_IGMP_SNOOPING_DEBUG					
					printk("mdb_del_by_port\n");
#endif
					if(mdst == mdb->mdst) /* remove first */
						{
						mdb->mdst=mdst->next_port;
						kfree(mdst);
						mdst=mdb->mdst;
						before=mdst;
						}
					else
						{
						before->next_port=mdst->next_port;
						kfree(mdst);
						mdst=before->next_port;
						}
					}
				else
					{
					before=mdst;
					mdst=mdst->next_port;
					}
				}		
			if(mdb->mdst == NULL)
				{
				pre_mdb=mdb;
				mdb=mdb->next_hash;
				__mhash_unlink(pre_mdb);
				kfree(pre_mdb);
				}
			else
				mdb = mdb->next_hash;
			}
		}
	read_unlock_bh(&br->mhash_lock);
	
}
#endif
#else /* End of Performance optimizations */ 

#include <linux/kernel.h>
#include <linux/spinlock.h>
#include <linux/if_bridge.h>
#include <asm/atomic.h>
#include <asm/uaccess.h>
#include "br_private.h"
/* frank,040202 */
#ifdef ASUS_IGMP_SNOOPING
#include <linux/inetdevice.h>
#include <linux/ip.h>
#include <asm/checksum.h>
#endif

static __inline__ unsigned long __timeout(struct net_bridge *br)
{
	unsigned long timeout;

	timeout = jiffies - br->ageing_time;
	if (br->topology_change)
		timeout = jiffies - br->forward_delay;

	return timeout;
}

static __inline__ int has_expired(struct net_bridge *br,
				  struct net_bridge_fdb_entry *fdb)
{
	if (!fdb->is_static &&
	    time_before_eq(fdb->ageing_timer, __timeout(br)))
		return 1;

	return 0;
}

static __inline__ void copy_fdb(struct __fdb_entry *ent, struct net_bridge_fdb_entry *f)
{
	memset(ent, 0, sizeof(struct __fdb_entry));
	memcpy(ent->mac_addr, f->addr.addr, ETH_ALEN);
	ent->port_no = f->dst?f->dst->port_no:0;
	ent->is_local = f->is_local;
	ent->ageing_timer_value = 0;
	if (!f->is_static)
		ent->ageing_timer_value = jiffies - f->ageing_timer;
}

static __inline__ int br_mac_hash(unsigned char *mac)
{
	unsigned long x;

	x = mac[0];
	x = (x << 2) ^ mac[1];
	x = (x << 2) ^ mac[2];
	x = (x << 2) ^ mac[3];
	x = (x << 2) ^ mac[4];
	x = (x << 2) ^ mac[5];

	x ^= x >> 8;

	return x & (BR_HASH_SIZE - 1);
}

static __inline__ void __hash_link(struct net_bridge *br,
				   struct net_bridge_fdb_entry *ent,
				   int hash)
{
	ent->next_hash = br->hash[hash];
	if (ent->next_hash != NULL)
		ent->next_hash->pprev_hash = &ent->next_hash;
	br->hash[hash] = ent;
	ent->pprev_hash = &br->hash[hash];
}

static __inline__ void __hash_unlink(struct net_bridge_fdb_entry *ent)
{
	*(ent->pprev_hash) = ent->next_hash;
	if (ent->next_hash != NULL)
		ent->next_hash->pprev_hash = ent->pprev_hash;
	ent->next_hash = NULL;
	ent->pprev_hash = NULL;
}



void br_fdb_changeaddr(struct net_bridge_port *p, unsigned char *newaddr)
{
	struct net_bridge *br;
	int i;

	br = p->br;
	write_lock_bh(&br->hash_lock);
	for (i=0;i<BR_HASH_SIZE;i++) {
		struct net_bridge_fdb_entry *f;

		f = br->hash[i];
		while (f != NULL) {
			if (f->dst == p && f->is_local) {
				__hash_unlink(f);
				memcpy(f->addr.addr, newaddr, ETH_ALEN);
				__hash_link(br, f, br_mac_hash(newaddr));
				write_unlock_bh(&br->hash_lock);
				return;
			}
			f = f->next_hash;
		}
	}
	write_unlock_bh(&br->hash_lock);
}

void br_fdb_cleanup(struct net_bridge *br)
{
	int i;
	unsigned long timeout;

	timeout = __timeout(br);

	write_lock_bh(&br->hash_lock);
	for (i=0;i<BR_HASH_SIZE;i++) {
		struct net_bridge_fdb_entry *f;

		f = br->hash[i];
		while (f != NULL) {
			struct net_bridge_fdb_entry *g;

			g = f->next_hash;
			if (!f->is_static &&
			    time_before_eq(f->ageing_timer, timeout)) {
				__hash_unlink(f);
				br_fdb_put(f);
			}
			f = g;
		}
	}
	write_unlock_bh(&br->hash_lock);
}

void br_fdb_delete_by_port(struct net_bridge *br, struct net_bridge_port *p)
{
	int i;

	write_lock_bh(&br->hash_lock);
	for (i=0;i<BR_HASH_SIZE;i++) {
		struct net_bridge_fdb_entry *f;

		f = br->hash[i];
		while (f != NULL) {
			struct net_bridge_fdb_entry *g;

			g = f->next_hash;
			if (f->dst == p) {
				__hash_unlink(f);
				br_fdb_put(f);
			}
			f = g;
		}
	}
	write_unlock_bh(&br->hash_lock);
}

struct net_bridge_fdb_entry *br_fdb_get(struct net_bridge *br, unsigned char *addr)
{
	struct net_bridge_fdb_entry *fdb;

	read_lock_bh(&br->hash_lock);
	fdb = br->hash[br_mac_hash(addr)];
	while (fdb != NULL) {

#if defined (CONFIG_MIPS_AVALANCHE_FAST_BRIDGE)

		if (  (*((short *)fdb->addr.addr) ==  *((short *)addr)) && ((*(short  *)(((char *)fdb->addr.addr + 2))) == (*(short  *)(addr + 2))) 
			&& ((*(short  *)(((char *)fdb->addr.addr + 4))) == (*(short  *)(addr + 4)))  )   {
#else
		if (!memcmp(fdb->addr.addr, addr, ETH_ALEN)) {
#endif
			if (!has_expired(br, fdb)) {
				atomic_inc(&fdb->use_count);
				read_unlock_bh(&br->hash_lock);
				return fdb;
			}

			read_unlock_bh(&br->hash_lock);
			return NULL;
		}

		fdb = fdb->next_hash;
	}

	read_unlock_bh(&br->hash_lock);
	return NULL;
}

void br_fdb_put(struct net_bridge_fdb_entry *ent)
{
	if (atomic_dec_and_test(&ent->use_count))
		kfree(ent);
}

int br_fdb_get_entries(struct net_bridge *br,
		       unsigned char *_buf,
		       int maxnum,
		       int offset)
{
	int i;
	int num;
	struct __fdb_entry *walk;

	num = 0;
	walk = (struct __fdb_entry *)_buf;

	read_lock_bh(&br->hash_lock);
	for (i=0;i<BR_HASH_SIZE;i++) {
		struct net_bridge_fdb_entry *f;

		f = br->hash[i];
		while (f != NULL && num < maxnum) {
			struct __fdb_entry ent;
			int err;
			struct net_bridge_fdb_entry *g;
			struct net_bridge_fdb_entry **pp; 

			if (has_expired(br, f)) {
				f = f->next_hash;
				continue;
			}

			if (offset) {
				offset--;
				f = f->next_hash;
				continue;
			}

			copy_fdb(&ent, f);

			atomic_inc(&f->use_count);
			read_unlock_bh(&br->hash_lock);
			err = copy_to_user(walk, &ent, sizeof(struct __fdb_entry));
			read_lock_bh(&br->hash_lock);

			g = f->next_hash;
			pp = f->pprev_hash;
			br_fdb_put(f);

			if (err)
				goto out_fault;

			if (g == NULL && pp == NULL)
				goto out_disappeared;

			num++;
			walk++;

			f = g;
		}
	}

 out:
	read_unlock_bh(&br->hash_lock);
	return num;

 out_disappeared:
	num = -EAGAIN;
	goto out;

 out_fault:
	num = -EFAULT;
	goto out;
}

static __inline__ void __fdb_possibly_replace(struct net_bridge_fdb_entry *fdb,
					      struct net_bridge_port *source,
					      int is_local)
{
	if (!fdb->is_static || is_local) {
		fdb->dst = source;
		fdb->is_local = is_local;
		fdb->is_static = is_local;
		fdb->ageing_timer = jiffies;
	}
}

void br_fdb_insert(struct net_bridge *br,
		   struct net_bridge_port *source,
		   unsigned char *addr,
		   int is_local)
{
	struct net_bridge_fdb_entry *fdb;
	int hash;

	hash = br_mac_hash(addr);

	write_lock_bh(&br->hash_lock);
	fdb = br->hash[hash];
	while (fdb != NULL) {
		if (!memcmp(fdb->addr.addr, addr, ETH_ALEN)) {
			__fdb_possibly_replace(fdb, source, is_local);
			write_unlock_bh(&br->hash_lock);
			return;
		}

		fdb = fdb->next_hash;
	}

	fdb = kmalloc(sizeof(*fdb), GFP_ATOMIC);
	if (fdb == NULL) {
		write_unlock_bh(&br->hash_lock);
		return;
	}

	memcpy(fdb->addr.addr, addr, ETH_ALEN);
	atomic_set(&fdb->use_count, 1);
	fdb->dst = source;
	fdb->is_local = is_local;
	fdb->is_static = is_local;
	fdb->ageing_timer = jiffies;

	__hash_link(br, fdb, hash);

	write_unlock_bh(&br->hash_lock);
}
/* frank,040202 */
#ifdef ASUS_IGMP_SNOOPING
static __inline__ void __mhash_link(struct net_bridge *br,
				   struct net_bridge_mdb_entry *ent,
				   int mhash)
{
	ent->next_hash = br->mhash[mhash];
	if (ent->next_hash != NULL)
		ent->next_hash->pprev_hash = &ent->next_hash;
	br->mhash[mhash] = ent;
	ent->pprev_hash = &br->mhash[mhash];
}

static __inline__ void __mhash_unlink(struct net_bridge_mdb_entry *ent)
{
	*(ent->pprev_hash) = ent->next_hash;
	if (ent->next_hash != NULL)
		ent->next_hash->pprev_hash = ent->pprev_hash;
	ent->next_hash = NULL;
	ent->pprev_hash = NULL;
}

struct net_bridge_mdb_entry *br_mdb_get(struct net_bridge *br, unsigned char *ip)
{
	struct net_bridge_mdb_entry *mdb;
	unsigned char addr[6]={0,0};

	read_lock_bh(&br->mhash_lock);
	memcpy(&addr[2],ip,4);
	mdb = br->mhash[br_mac_hash(addr)];
	while (mdb != NULL) {
		if (mdb->addr==*((unsigned int *)ip)) {				
			read_unlock_bh(&br->mhash_lock);
			return mdb;
		}
		mdb = mdb->next_hash;
	}
	read_unlock_bh(&br->mhash_lock);
	return NULL;
}
void br_mdb_insert(struct net_bridge *br,
		   struct net_bridge_port *source,
		   unsigned char *ip)
{
	struct net_bridge_mdb_entry *mdb;
	struct net_bridge_mport	 *mdst;
	unsigned char addr[6]={0,0};
	int mhash;

	memcpy(&addr[2],ip,4);
	mhash = br_mac_hash(addr);

	write_lock_bh(&br->mhash_lock);
	mdb = br->mhash[mhash];
	while (mdb != NULL) {
		if (mdb->addr==*((unsigned int *) ip)) {
			/* check port */
			mdst=mdb->mdst;
			while(mdst != NULL)
				{
				if(mdst->dst == source)
					{
					/* old ones,reset aging timer */
					mdst->ageing_timer=jiffies;
#ifdef ASUS_IGMP_SNOOPING_DEBUG					
					printk("[%u]reset timer from [%s] for [01:00:5E:%02X:%02X:%02X]\n",jiffies/100,source->dev->name,addr[3],addr[4],addr[5]);
#endif
					break;
					}
				else
					mdst=mdst->next_port;
				}
			/* not found bridge port in mfdb,insert port to mdb */
			if(mdst == NULL)
				{
				mdst=kmalloc(sizeof(*mdst), GFP_ATOMIC);			
				if(mdst == NULL)
					{
					write_unlock_bh(&br->mhash_lock);
					return;
					}
				mdst->ageing_timer=jiffies;
				mdst->dst=source;
				mdst->next_port=mdb->mdst;
				mdb->mdst=mdst;
#ifdef ASUS_IGMP_SNOOPING_DEBUG					
				printk("[%u]new mdst  from [%s] for [01:00:5E:%02X:%02X:%02X]\n",jiffies/100,source->dev->name,addr[3],addr[4],addr[5]);
#endif
				}
			write_unlock_bh(&br->mhash_lock);
			return;
		}

		mdb = mdb->next_hash;
	}

	mdb = kmalloc(sizeof(*mdb), GFP_ATOMIC);
	mdst=kmalloc(sizeof(*mdst), GFP_ATOMIC);
	
	if (mdb == NULL || mdst == NULL) {
		if(mdb != NULL)
			kfree(mdb);
		if(mdst != NULL)
			kfree(mdst);
		write_unlock_bh(&br->mhash_lock);
		return;
	}
	mdst->ageing_timer=jiffies;
	mdst->next_port=NULL;
	mdst->dst=source;
	mdb->addr=*((unsigned int *)ip);
	/* copy igmp packet */	
		

	atomic_set(&mdb->use_count, 1);
	mdb->mdst = mdst;

	__mhash_link(br, mdb, mhash);
#ifdef ASUS_IGMP_SNOOPING_DEBUG					
	printk("[%u]mew mdb  from [%s] for [01:00:5E:%02X:%02X:%02X]\n",jiffies/100,source->dev->name,addr[3],addr[4],addr[5]);
#endif
	write_unlock_bh(&br->mhash_lock);
}

void br_mdb_leave_group(struct net_bridge *br,
		   struct net_bridge_port *source,
		   unsigned char *group)
{
	struct net_bridge_mdb_entry *mdb;
	struct net_bridge_mport	 *mdst,*before;
	unsigned char addr[6]={0x00,0x00};

	memcpy(&addr[2],group,4);
#ifdef ASUS_IGMP_SNOOPING_DEBUG					
	printk("Receive IGMP Leave Group[01:00:5e:%02X:%02X:%02X] from port[%s]\n",addr[3],addr[4],addr[5],source->dev->name);
#endif	
	read_lock_bh(&br->mhash_lock);
	mdb = br->mhash[br_mac_hash(addr)];
	while (mdb != NULL) {
		if (mdb->addr==*((unsigned int *)group)) {
			mdst=mdb->mdst;
			before=mdst;
			while(mdst != NULL)
				{
				if(mdst->dst == source)
					{
					if(mdst == mdb->mdst) /* remove first */
						{
						mdb->mdst=mdst->next_port;
						kfree(mdst);
						mdst=mdb->mdst;
						before=mdst;
						}
					else
						{
						before->next_port=mdst->next_port;
						kfree(mdst);
						mdst=before->next_port;
						}
#ifdef ASUS_IGMP_SNOOPING_DEBUG					
					printk("Find Entry and delete\n\n");
#endif
					break;
					}
				else
					{
					before=mdst;
					mdst=mdst->next_port;
					}
				}
			if(mdb->mdst == NULL)
				{
				__mhash_unlink(mdb);
					{
#ifdef ASUS_IGMP_SNOOPING_DEBUG					
					printk("clean MDB\n");
#endif
					kfree(mdb);
					mdb=NULL;
					}
				}
				
			read_unlock_bh(&br->mhash_lock);
			return;
		}

		mdb = mdb->next_hash;
	}

	read_unlock_bh(&br->mhash_lock);
}

static unsigned char igmp_query_pkt[60]={
	0x01,0x00,0x5e,0x00,0x00,0x01,0x00,0x00,0x00,0x00,0x00,0x00,0x08,0x00,0x45,0xc0,
	0x00,0x1c,0xe5,0xb4,0x00,0x00,0x01,0x02,0xea,0xc0,0xc0,0xa8,0x48,0x01,0xe0,0x00,
	0x00,0x01,0x11,0x64,0xee,0x9b,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00};
void br_send_igmp_query(struct net_bridge_port *port)
{
	struct net_device *dev;
	struct sk_buff *skb;
	int size;
	struct in_ifaddr *ifa=NULL;
	struct iphdr	*iph;

	/* get ip address of br */
	if(port->br->dev.ip_ptr != NULL)
		ifa = ((struct in_device*)(port->br->dev.ip_ptr))->ifa_list;
	
	
	size = 60;
	dev = port->dev;

	if ((skb = dev_alloc_skb(size)) == NULL) {
		printk(KERN_INFO "br: memory squeeze!\n");
		return;
	}
#ifdef ASUS_IGMP_SNOOPING_DEBUG					
	printk("[%u]send IGMP Query:port[%s] \n\n",jiffies/HZ,port->dev->name);
#endif
	skb->dev = dev;
	skb->protocol = htons(ETH_P_IP);
	skb->mac.raw = skb_put(skb, size);
	memcpy(skb->mac.raw, igmp_query_pkt, size);
	memcpy(skb->mac.raw+ETH_ALEN, dev->dev_addr, ETH_ALEN);

	if(ifa != NULL)
		{
		(unsigned char *)iph=skb->mac.raw+14;
		iph->saddr=ifa->ifa_address;
		iph->check=0;
		iph->check=ip_fast_csum((unsigned char *) iph, iph->ihl);
		}
	dev_queue_xmit(skb);


}

int br_mdb_get_entries(struct net_bridge *br,
		       unsigned char *_buf,
		       int maxnum,
		       int offset)
{
	struct net_bridge_mdb_entry *mdb;
	struct net_bridge_mport	 *mdst;
	struct __mdb_entry *data,ent;
	int index=0,i;

	data=(struct __mdb_entry *)_buf;

	read_lock_bh(&br->mhash_lock);
		
	for(i=0;i<BR_HASH_SIZE;i++)
		{
		mdb = br->mhash[i];
		while (mdb != NULL) 
			{
			mdst=mdb->mdst;
			while(mdst != NULL) 
				{
				ent.ip_addr=mdb->addr;
				ent.ageing_timer_value=(jiffies-mdst->ageing_timer)/HZ;
				ent.port_no=mdst->dst->port_no;
				copy_to_user(&data[index], &ent, sizeof(struct __mdb_entry));
				index++;
				if(index >= maxnum)
					{
					read_unlock_bh(&br->mhash_lock);
					return index;
					}
				mdst=mdst->next_port;
				}				
			mdb = mdb->next_hash;
			}
		}
	read_unlock_bh(&br->mhash_lock);
	return index;
}

void br_mdb_cleanup(struct net_bridge *br)
{
	struct net_bridge_mdb_entry *mdb,*pre_mdb;
	struct net_bridge_mport	 *mdst,*pre_mdst;
	int i;

	read_lock_bh(&br->mhash_lock);
	for(i=0;i<BR_HASH_SIZE;i++)
		{
		mdb = br->mhash[i];
		while (mdb != NULL) 
			{
			mdst=mdb->mdst;
			while(mdst != NULL) 
				{
				pre_mdst=mdst;
				mdst=mdst->next_port;
				kfree(pre_mdst);
				}		
			pre_mdb=mdb;
			mdb=mdb->next_hash;
			__mhash_unlink(pre_mdb);
			kfree(pre_mdb);
			}
		}
	read_unlock_bh(&br->mhash_lock);
}

void br_mdb_delete_by_port(struct net_bridge *br, struct net_bridge_port *p)
{
	struct net_bridge_mdb_entry *mdb,*pre_mdb;
	struct net_bridge_mport	 *mdst,*before;
	int i;

	read_lock_bh(&br->mhash_lock);
	for(i=0;i<BR_HASH_SIZE;i++)
		{
		mdb = br->mhash[i];
		while (mdb != NULL) 
			{
			mdst=mdb->mdst;
			before=mdst;
			while(mdst != NULL) 
				{
				if(mdst->dst == p)
					{
#ifdef ASUS_IGMP_SNOOPING_DEBUG					
					printk("mdb_del_by_port\n");
#endif
					if(mdst == mdb->mdst) /* remove first */
						{
						mdb->mdst=mdst->next_port;
						kfree(mdst);
						mdst=mdb->mdst;
						before=mdst;
						}
					else
						{
						before->next_port=mdst->next_port;
						kfree(mdst);
						mdst=before->next_port;
						}
					}
				else
					{
					before=mdst;
					mdst=mdst->next_port;
					}
				}		
			if(mdb->mdst == NULL)
				{
				pre_mdb=mdb;
				mdb=mdb->next_hash;
				__mhash_unlink(pre_mdb);
				kfree(pre_mdb);
				}
			else
				mdb = mdb->next_hash;
			}
		}
	read_unlock_bh(&br->mhash_lock);
	
}
#endif

#endif /* TI_SLOW_PATH */
